from entities.descriptors import CommandDescriptor, PayloadFieldEnum
from entities.descriptors import PayloadFieldU8, PayloadFieldU16, PayloadFieldU32
from switches import SwitchList
from analogvariables import AnalogVariableList
from digitalvariables import DigitalVariableList
from userpins import UserPinList
from firmwareconfig import FirmwareConfig as fwcfg


class SwitchEnumField(PayloadFieldEnum):
    """A class to create enum relative to the switches"""

    def __init__(self, switch_list):
        assert isinstance(switch_list, SwitchList)
        PayloadFieldEnum.__init__(
            self,
            name='SWITCHNUM',
            description='Switch number',
            index_list=[itm.enum_index for itm in switch_list.entries()],
            value_list=[itm.name for itm in switch_list.entries()])


class UserPinEnumField(PayloadFieldEnum):
    """A class to create enum relative to the board user pins"""

    def __init__(self, user_pin_list):
        assert isinstance(user_pin_list, UserPinList)
        PayloadFieldEnum.__init__(
            self,
            name='PINNUM',
            description='User pin number',
            index_list=[itm.enum_index for itm in user_pin_list.entries()],
            value_list=[itm.name for itm in user_pin_list.entries()])


class AlarmEnumField(PayloadFieldEnum):

    def __init__(self, analog_variable_list, digital_variable_list, analog_only=False, append_all_selector=False, append_ored_selector=False):
        assert isinstance(analog_variable_list, AnalogVariableList)
        assert isinstance(digital_variable_list, DigitalVariableList)
        index_list = list()
        value_list = list()
        enum_name = 'ALARM_NUMBER'
        for variable in analog_variable_list.entries():
            if variable.alarm_slow:
                index_list.append(variable.alarm_slow.enum_index)
                value_list.append('ALARM_SLOW_' + variable.name)
            if variable.alarm_fast:
                index_list.append(variable.alarm_fast.enum_index)
                value_list.append('ALARM_FAST_' + variable.name)
        if not analog_only:
            for variable in digital_variable_list.entries():
                if variable.alarm:
                    index_list.append(variable.alarm.enum_index)
                    value_list.append('ALARM_' + variable.name)
        else:
            enum_name += '_ANALOG'
        if append_all_selector:
            enum_name += '_ALL'
            index_list.append(max(index_list)+1)
            value_list.append('ALARMS_ALL')
        if append_ored_selector:
            enum_name += '_EXT'
            index_list.append(max(index_list)+1)
            value_list.append('ALARMS_ALL_ORED')

        PayloadFieldEnum.__init__(
            self,
            name=enum_name,
            description='Alarm number',
            index_list=index_list,
            value_list=value_list
        )


class VariableEnumField(PayloadFieldEnum):

    def __init__(self, analog_variable_list, digital_variable_list, analog_only=False, append_all_selector=False):
        assert isinstance(analog_variable_list, AnalogVariableList)
        assert isinstance(digital_variable_list, DigitalVariableList)
        index_list = list()
        value_list = list()
        enum_name = 'VARIABLE_NUMBER'
        for variable in analog_variable_list.entries():
            index_list.append(variable.enum_index)
            value_list.append(variable.name)
        if not analog_only:
            for variable in digital_variable_list.entries():
                index_list.append(variable.enum_index)
                value_list.append(variable.name)
        else:
            enum_name += '_ANALOG'
        if append_all_selector:
            enum_name += '_ALL'
            index_list.append(max(index_list)+1)
            value_list.append('ALL_VARIABLES')

        PayloadFieldEnum.__init__(
            self,
            name=enum_name,
            description='Variable number',
            index_list=index_list,
            value_list=value_list
        )


class VariableAveragingPrescalerEnumField(PayloadFieldEnum):
    """A class to create enum relative to the averaging functionality"""

    def __init__(self):
        idxs = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
        PayloadFieldEnum.__init__(
            self,
            name='AVERAGINGPRESCALER',
            description='Variable averaging prescaler',
            index_list=idxs,
            value_list=[str(idx) for idx in idxs])


class SwitchStateField(PayloadFieldEnum):
    """A class to create enum relative to board switch state"""

    def __init__(self, for_request=False):
        if for_request:
            PayloadFieldEnum.__init__(
                self,
                name='SWITCHSTATE_NC',
                description='Switch state',
                index_list=[0, 1, 2],
                value_list=['OPEN', 'CLOSED', 'NO_CHANGE'])
        else:
            PayloadFieldEnum.__init__(
                self,
                name='SWITCHSTATE',
                description='Switch state',
                index_list=[0, 1],
                value_list=['OPEN', 'CLOSED'])


class EnableStateField(PayloadFieldEnum):

    def __init__(self, for_request=False):
        if for_request:
            PayloadFieldEnum.__init__(
                self,
                name='ENABLESTATE_NC',
                description='Enable state',
                index_list=[0, 1, 2],
                value_list=['DISABLED', 'ENABLED', 'NO_CHANGE'])
        else:
            PayloadFieldEnum.__init__(
                self,
                name='ENABLESTATE',
                description='Enable state',
                index_list=[0, 1],
                value_list=['DISABLED', 'ENABLED'])

class EnableStateFieldNamed(PayloadFieldEnum):

    def __init__(self, description):
        PayloadFieldEnum.__init__(
            self,
            name='ENABLESTATE',
            description=description,
            index_list=[0, 1],
            value_list=['DISABLED', 'ENABLED'])


class FiredStateField(PayloadFieldEnum):

    def __init__(self):
        PayloadFieldEnum.__init__(
            self,
            name='FIREDSTATE',
            description='Fired state',
            index_list=[0, 1],
            value_list=['NO', 'YES'])


class FiredStateFieldNamed(PayloadFieldEnum):

    def __init__(self, description):
        PayloadFieldEnum.__init__(
            self,
            name='FIREDSTATE',
            description=description,
            index_list=[0, 1],
            value_list=['NO', 'YES'])


class FlagStateField(PayloadFieldEnum):

    def __init__(self, name=None, description=None):
        if name is None:
            name = 'FLAGSTATE'
        if description is None:
            description = 'Status'
        PayloadFieldEnum.__init__(
            self,
            name=name,
            description=description,
            index_list=[0, 1],
            value_list=['LOW', 'HIGH'])


class UserPinStateField(PayloadFieldEnum):

    def __init__(self, name=None, description=None):
        if name is None:
            name = 'PINSTATE'
        if description is None:
            description = 'Status'
        PayloadFieldEnum.__init__(
            self,
            name=name,
            description=description,
            index_list=[0, 1],
            value_list=['LOW', 'HIGH'])


class CommandList:
    """A class to define the list of board commands"""

    # this class can be actually created only once
    _ready = False

    def __init__(self, switch_list, analog_variable_list, digital_variable_list, user_pin_list):
        assert self._ready is False, "Class {} should be created only once!".format(self.__class__.__name__)
        assert isinstance(switch_list, SwitchList)
        assert isinstance(analog_variable_list, AnalogVariableList)
        assert isinstance(digital_variable_list, DigitalVariableList)
        assert isinstance(user_pin_list, UserPinList)
        self.switch_list = switch_list
        self.analog_variable_list = analog_variable_list
        self.digital_variable_list = digital_variable_list
        self.user_pin_list = user_pin_list
        self._create()
        self._ready = True

    def count(self):
        return len(self.vars)

    def entries(self):
        return self.vars

    def _create(self):
        self.vars = list()
        self._create_field_types()
        self._create_switch_commands()
        self._create_alarm_commands()
        self._create_variable_commands()
        self._create_utility_commands()
        self._create_rescue_control_commands()
        self._create_debug_commands()
        self._consolidate_list()

    def _consolidate_list(self):
        command_code_start = 0x40  # to avoid clashing with previus version commands
        command_code_current = command_code_start
        command_code_usedlist = []
        for cmd in self.vars:
            if cmd.request_code is None:
                assert cmd.response_code is None
                assert command_code_current not in command_code_usedlist
                assert (command_code_current+1) not in command_code_usedlist
                cmd.request_code = command_code_current
                cmd.response_code = command_code_current+1
                command_code_current += 2
            else:
                if cmd.response_code is None:
                    assert (cmd.request_code + 1) not in command_code_usedlist
                    cmd.response_code = cmd.requset_code + 1
            assert cmd.request_code not in command_code_usedlist, "Command {}".format(cmd.name)
            command_code_usedlist.append(cmd.request_code)
            assert cmd.response_code not in command_code_usedlist or cmd.implement is False, "Command {}".format(cmd.name)
            command_code_usedlist.append(cmd.response_code)

    def _create_field_types(self):
        # switch enumeration
        self.switch_enum_field = SwitchEnumField(self.switch_list)

        # variables enum fields
        self.variable_enum_field = VariableEnumField(self.analog_variable_list, self.digital_variable_list)
        self.variable_enum_field_all = VariableEnumField(self.analog_variable_list, self.digital_variable_list, append_all_selector=True)

        # variables value fields
        self.variable_value_field = PayloadFieldU16(name='VALUE', description='Last acquired sensor value (0/1 for digital sensors)')
        self.variable_meanvalue_field = PayloadFieldU16(name='MEANVALUE', description='Sensor mean value (0 for digital sensors)')
        self.variable_offset_field = PayloadFieldU16(name='OFFSET', description='Sensor offset value (acquired at start-up, reports zero for digital sensors)')
        self.variable_maxvalue_field = PayloadFieldU16(name='MAXVALUE', description='Sensor maximum value since last reset (0/1 for digital sensors)')

        # some variables command fields generated collections
        self.variable_getall_values_respfields = self.__util_gen_variable_getall_response_payload('values')
        self.variable_getall_meanvalues_respfields = self.__util_gen_variable_getall_response_payload('means')
        self.variable_getall_offsets_respfields = self.__util_gen_variable_getall_response_payload('offsets')
        self.variable_getall_maxvalues_respfields = self.__util_gen_variable_getall_response_payload('maxvalues')
        self.variable_meanvalue_prescaler_field = VariableAveragingPrescalerEnumField()

        # alarms enum fields
        self.alarm_enum_field = AlarmEnumField(self.analog_variable_list, self.digital_variable_list)
        self.alarm_enum_field_all = AlarmEnumField(self.analog_variable_list, self.digital_variable_list, append_all_selector=True)
        self.alarm_enum_field_analog = AlarmEnumField(self.analog_variable_list, self.digital_variable_list, analog_only=True)
        self.alarm_enum_field_ored = AlarmEnumField(self.analog_variable_list, self.digital_variable_list, append_ored_selector=True)

        # some alarms commands fields generated collections
        self.alarm_getall_enable_respfields = self.__util_gen_alarm_getall_response_payload('enable')
        self.alarm_getall_maxcount_respfields = self.__util_gen_alarm_getall_response_payload('maxcount')
        self.alarm_getall_fired_respfields = self.__util_gen_alarm_getall_response_payload('fired')
        self.alarm_getall_firecount_respfields = self.__util_gen_alarm_getall_response_payload('firecount')

        # alarms value fields
        self.alarm_timeout_value_field = PayloadFieldU16(name='TIMEOUT', description='Timeout value')
        self.alarm_maxcount_value_field = PayloadFieldU16(name='MAXCOUNT', description='Max count value')
        self.alarm_firecount_value_field = PayloadFieldU16(name='FIRECOUNT', description='Fire count value')
        self.alarm_threshold_value_field = PayloadFieldU16(name='THRESHOLD', description='Threshold value')

        # rescue procedure fields
        self.rescue_timeout_value_field = PayloadFieldU16(name='RESCUE_TIMEOUT', description='Timeout value')

        # user pin related enumeration
        self.user_pin_enum_field = UserPinEnumField(self.user_pin_list)

    def _create_switch_commands(self):

        description = 'Control one of the switches or enable lines of the board. Can be used to get the current switch status without changing it (see request payload field status)'
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='SWITCH_CONTROL',
                request_code=None,
                response_code=None,
                description=description,
                request_payload=[self.switch_enum_field, SwitchStateField(for_request=True)],
                response_payload=[self.switch_enum_field, SwitchStateField()]
            )
        )

    def _create_alarm_commands(self):

        ################################################################################################################
        # Alarm enable/disable command
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='ALARM_ENABLE',
                request_code=None,
                response_code=None,
                description='Enable or disable one of the alarms. Can be used to get the current enable status without changing it (see request payload field)',
                request_payload=[self.alarm_enum_field, EnableStateField(for_request=True)],
                response_payload=[self.alarm_enum_field, EnableStateField()]
            )
        )

        ################################################################################################################
        # Append command: Set timeout command
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='ALARM_TIMEOUT_SET',
                request_code=None,
                response_code=None,
                description='Set the timeout of one alarm',
                request_payload=[self.alarm_enum_field, self.alarm_timeout_value_field],
                response_payload=[]
            )
        )

        ################################################################################################################
        # Append command: Get timeout command
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='ALARM_TIMEOUT_GET',
                request_code=None,
                response_code=None,
                description='Get the timeout value of one alarm',
                request_payload=[self.alarm_enum_field],
                response_payload=[self.alarm_enum_field, self.alarm_timeout_value_field]
            )
        )

        ################################################################################################################
        # Append command: Get the max timeout counter command
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='ALARM_MAXCOUNT_GET',
                request_code=None,
                response_code=None,
                description='Get the maximum counter value reached by one of the alarms (use command ALARM_MAXCOUNT_RESET to reset it to zero)',
                request_payload=[self.alarm_enum_field],
                response_payload=[self.alarm_enum_field, self.alarm_maxcount_value_field]
            )
        )

        ################################################################################################################
        # Append command: Max timeout counter reset command
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='ALARM_MAXCOUNT_RESET',
                request_code=None,
                response_code=None,
                description='Reset the maximum counter value of one alarm (or all alarms)',
                request_payload=[self.alarm_enum_field_all],
                response_payload=[]
            )
        )

        ################################################################################################################
        # Append command: Set alarm threshold command (for analog sensor alarms only)
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='ALARM_THRESHOLD_SET',
                request_code=None,
                response_code=None,
                description='Set the threshold value of one analog alarm',
                request_payload=[self.alarm_enum_field_analog, self.alarm_threshold_value_field],
                response_payload=[]
            )
        )

        ################################################################################################################
        # Append command: Get alarm threshold command (for analog sensor alarms only)
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='ALARM_THRESHOLD_GET',
                request_code=None,
                response_code=None,
                description='Get the threshold value of one analog alarm',
                request_payload=[self.alarm_enum_field_analog],
                response_payload=[self.alarm_enum_field_analog, self.alarm_threshold_value_field]
            )
        )

        ################################################################################################################
        # Append command: Get alarm fired command
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='ALARM_FIRED_GET',
                request_code=None,
                response_code=None,
                description='Get the has-fired flag of one alarm (or all alarms OR-ed)',
                request_payload=[self.alarm_enum_field_ored],
                response_payload=[self.alarm_enum_field_ored, FiredStateField()]
            )
        )

        ################################################################################################################
        # Append command: Alarm fired reset command
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='ALARM_FIRED_RESET',
                request_code=None,
                response_code=None,
                description='Reset the has-fired flag of one alarm (or all alarms)',
                request_payload=[self.alarm_enum_field_all],
                response_payload=[]
            )
        )

        ################################################################################################################
        # Append command: Get alarm firecount command
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='ALARM_FIRECOUNT_GET',
                request_code=None,
                response_code=None,
                description='Get the current fire counter of one alarm',
                request_payload=[self.alarm_enum_field],
                response_payload=[self.alarm_enum_field, self.alarm_firecount_value_field]
            )
        )

        ################################################################################################################
        # Append command: Get alarm enable getall command
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='ALARM_ENABLE_GETALL',
                request_code=None,
                response_code=None,
                description='Get the current enable status of all alarms',
                request_payload=[],
                response_payload=self.alarm_getall_enable_respfields
            )
        )

        ################################################################################################################
        # Append command: Get alarm maxcount getall command
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='ALARM_MAXCOUNT_GETALL',
                request_code=None,
                response_code=None,
                description='Get the current maxcount value of all alarms',
                request_payload=[],
                response_payload=self.alarm_getall_maxcount_respfields
            )
        )

        ################################################################################################################
        # Append command: Get alarm fired getall command
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='ALARM_FIRED_GETALL',
                request_code=None,
                response_code=None,
                description='Get the current fired state of all alarms',
                request_payload=[],
                response_payload=self.alarm_getall_fired_respfields
            )
        )

        ################################################################################################################
        # Append command: Get alarm firecount getall command
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='ALARM_FIRECOUNT_GETALL',
                request_code=None,
                response_code=None,
                description='Get the current firecount value of all alarms',
                request_payload=[],
                response_payload=self.alarm_getall_firecount_respfields
            )
        )


    def __util_gen_variable_getall_response_payload(self, selection):
        """Utility function"""
        fields = list()
        for variable in self.analog_variable_list.entries():
            if selection == 'values':
                name = variable.name + '_VALUE'
                description = 'Variable ' + variable.name + ' last acquired value'
            elif selection == 'means':
                name = variable.name + '_MEAN'
                description = 'Variable ' + variable.name + ' mean'
            elif selection == 'offsets':
                name = variable.name + '_OFFSET'
                description = 'Variable ' + variable.name + ' offset'
            elif selection == 'maxvalues':
                name = variable.name + '_MAXVALUE'
                description = 'Variable ' + variable.name + ' max value'
            else:
                assert False
            fields.append(
                PayloadFieldU16(
                    name=name,
                    description=description
                )
            )
        # if requested fields are not offsets, then append the fields for the digital sensors
        if not selection == 'offsets':
            for variable in self.digital_variable_list.entries():
                if selection != 'means':
                    fields.append(
                        FlagStateField(
                            name=variable.name + '_STATE',
                            description=variable.name + ' status'
                        )
                    )
        return fields

    def __util_gen_alarm_getall_response_payload(self, selection):
        """Utility function"""

        # list of all alarms
        alarms = list()
        alarm_names = list()
        for analog_variable in self.analog_variable_list.entries():
            if analog_variable.alarm_slow:
                alarms.append(analog_variable.alarm_slow)
                alarm_names.append(analog_variable.name + '_ALARM_SLOW')
            if analog_variable.alarm_fast:
                alarms.append(analog_variable.alarm_fast)
                alarm_names.append(analog_variable.name + '_ALARM_FAST')
        for digital_variable in self.digital_variable_list.entries():
            if digital_variable.alarm:
                alarms.append(digital_variable.alarm)
                alarm_names.append(digital_variable.name + '_ALARM')

        # populate list of fields
        fields = list()
        for alarm_idx in range(len(alarms)):
            alarm = alarms[alarm_idx]
            alarm_name = alarm_names[alarm_idx]
            if selection == 'enable':
                # name = alarm_name + '_ENABLE'
                description = 'Alarm ' + alarm_name + ' enable state'
                fields.append(EnableStateFieldNamed(description=description))
            elif selection == 'maxcount':
                name = alarm_name + '_MAXCOUNT'
                description = 'Alarm ' + alarm_name + ' maxcount value'
                fields.append(PayloadFieldU16(name=name, description=description))
            elif selection == 'fired':
                # name = alarm_name + '_FIRED'
                description = 'Alarm ' + alarm_name + ' fired state'
                fields.append(FiredStateFieldNamed(description=description))
            elif selection == 'firecount':
                name = alarm_name + '_FIRECOUNT'
                description = 'Alarm ' + alarm_name + ' firecount value'
                fields.append(PayloadFieldU16(name=name, description=description))
            else:
                assert False

        return fields

    def _create_variable_commands(self):

        ################################################################################################################
        # Append command: Get one sensor current value, offset value, and max value
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='SENSOR_GET_SINGLE',
                request_code=None,
                response_code=None,
                description='Get one sensor\'s current value, offset value, max value, and mean value',
                request_payload=[self.variable_enum_field],
                response_payload=[self.variable_enum_field, self.variable_value_field, self.variable_offset_field, self.variable_maxvalue_field, self.variable_meanvalue_field]
            )
        )

        ################################################################################################################
        # Append command: Reset one sensor max value
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='SENSOR_MAXVALUE_RESET',
                request_code=None,
                response_code=None,
                description='Reset the max value of one sensor (or all sensors)',
                request_payload=[self.variable_enum_field_all],
                response_payload=[]
            )
        )

        ################################################################################################################
        # Append command: Get all sensors current values
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='SENSOR_VALUES_GETALL',
                request_code=None,
                response_code=None,
                description='Get all sensors\' values',
                request_payload=[],
                response_payload=self.variable_getall_values_respfields
            )
        )

        ################################################################################################################
        # Append command: Get all sensors mean values
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='SENSOR_AVERAGE_GETALL',
                request_code=None,
                response_code=None,
                description='Get all sensors\' mean values',
                request_payload=[],
                response_payload=self.variable_getall_meanvalues_respfields
            )
        )

        ################################################################################################################
        # Append command: Get all sensors offset values
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='SENSOR_OFFSETS_GETALL',
                request_code=None,
                response_code=None,
                description='Get all sensors\' offsets',
                request_payload=[],
                response_payload=self.variable_getall_offsets_respfields
            )
        )

        ################################################################################################################
        # Append command: Get all sensors max values
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='SENSOR_MAXVALUES_GETALL',
                request_code=None,
                response_code=None,
                description='Get all sensors\' max values',
                request_payload=[],
                response_payload=self.variable_getall_maxvalues_respfields
            )
        )

        ################################################################################################################
        # Append command: Set mean calculator prescale value
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='SENSOR_AVERAGING_PRESCALER_SET',
                request_code=None,
                response_code=None,
                description='Set the value of the prescaler used to compute sensors\' averages',
                request_payload=[self.variable_meanvalue_prescaler_field],
                response_payload=[]
            )
        )

        ################################################################################################################
        # Append command: Get mean calculator prescale value
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='SENSOR_AVERAGING_PRESCALER_GET',
                request_code=None,
                response_code=None,
                description='Get the value of the prescaler used to compute sensors\' averages',
                request_payload=[],
                response_payload=[self.variable_meanvalue_prescaler_field]
            )
        )

    def _create_utility_commands(self):

        # ------------------------------------------------------------------------------------------------------------------
        # Append command: Time Since reset
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='BOARDTIME',
                request_code=None,
                response_code=None,
                description='Time since last reset',
                request_payload=[],
                response_payload=[
                    PayloadFieldU32(
                        name='SECONDS',
                        description='Seconds since reset'
                    )]
            )
        )

        # ------------------------------------------------------------------------------------------------------------------
        # Append command: Time Since last command
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='LASTCOMMANDTIME',
                request_code=None,
                response_code=None,
                description='Time since last command',
                request_payload=[],
                response_payload=[
                    PayloadFieldU32(
                        name='SECONDS',
                        description='Seconds since last command'
                    )]
            )
        )

        # ------------------------------------------------------------------------------------------------------------------
        # Append command: Versioning info
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='VERSION',
                request_code=None,
                response_code=None,
                description='Report embedded code versioning information',
                request_payload=[],
                response_payload=[
                    PayloadFieldU32(
                        name='FIRMWARE_TARGET',
                        description='Target of the firmware'
                    ),
                    PayloadFieldU16(
                        name='COMMAND_TABLE_VERSION',
                        description='Version of the current command table'
                    ),
                    PayloadFieldU16(
                        name='TEMPLATE_GENERATION_YEAR',
                        description='Template generation year'
                    ),
                    PayloadFieldU8(
                        name='TEMPLATE_GENERATION_MONTH',
                        description='Template generation month'
                    ),
                    PayloadFieldU8(
                        name='TEMPLATE_GENERATION_DAY',
                        description='Template generation day'
                    ),
                    PayloadFieldU8(
                        name='TEMPLATE_GENERATION_HOUR',
                        description='Template generation hour'
                    ),
                    PayloadFieldU8(
                        name='TEMPLATE_GENERATION_MINUTE',
                        description='Template generation minute'
                    ),
                    PayloadFieldU8(
                        name='TEMPLATE_GENERATION_SECOND',
                        description='Template generation second'
                    )]
            )
        )

        # ------------------------------------------------------------------------------------------------------------------
        # Append command: Get last error recorded
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='LASTERROR',
                request_code=None,
                response_code=None,
                description='Return the last occurred error, and reset it to NOERR value',
                request_payload=[],
                response_payload=[
                    PayloadFieldU8(
                        name='ERRCODE',
                        description='Error code'
                    )]
            )
        )

    def _create_rescue_control_commands(self):

        # ------------------------------------------------------------------------------------------------------------------
        # Append command: commands related to rescuing procedure
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='RESCUE_ENABLE',
                request_code=None,
                response_code=None,
                description='Enable or disable auto-rescue (or read enable status)',
                request_payload=[EnableStateField(for_request=True)],
                response_payload=[EnableStateField()]
            )
        )

        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='RESCUE_STATUS_GET',
                request_code=None,
                response_code=None,
                description='Get the auto-rescue current status (fired or not)',
                request_payload=[],
                response_payload=[FiredStateField()]
            )
        )

        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='RESCUE_STATUS_RESET',
                request_code=None,
                response_code=None,
                description='Reset the auto-rescue current status',
                request_payload=[],
                response_payload=[]
            )
        )

        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='RESCUE_TIMEOUT_SET',
                request_code=None,
                response_code=None,
                description='Set the auto-rescue timeout value',
                request_payload=[self.rescue_timeout_value_field],
                response_payload=[]
            )
        )

        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='RESCUE_TIMEOUT_GET',
                request_code=None,
                response_code=None,
                description='Get the auto-rescue timeout value',
                request_payload=[],
                response_payload=[self.rescue_timeout_value_field]
            )
        )

    def _create_debug_commands(self):

        # ------------------------------------------------------------------------------------------------------------------
        # Append command: User pins commands

        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='USER_PIN_CONTROL',
                request_code=None,
                response_code=None,
                description='Control one user pin',
                request_payload=[self.user_pin_enum_field, UserPinStateField()],
                response_payload=[]
            )
        )

        # ------------------------------------------------------------------------------------------------------------------
        # Append command: Echo command (1 byte payload versions)
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='ECHO1',
                request_code=None,
                response_code=None,
                description='Echo (1 byte)',
                request_payload=[PayloadFieldU8(name='VALUE', description='Echoed byte')],
                response_payload=[PayloadFieldU8(name='VALUE', description='Echoed byte')]
            )
        )

        # ------------------------------------------------------------------------------------------------------------------
        # Append command: Echo command (4 bytes)
        _echo4_payload = list()
        for k in range(4):
            _echo4_payload.append(PayloadFieldU8(name='VALUE_' + str(k), description='Echoed byte'))
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='ECHO4',
                request_code=None,
                response_code=None,
                description='Echo (4 bytes)',
                request_payload=_echo4_payload,
                response_payload=_echo4_payload
            )
        )

        # ------------------------------------------------------------------------------------------------------------------
        # Append command: Echo command (32 bytes payload versions)
        _echo32_payload = list()
        for k in range(32):
            _echo32_payload.append(PayloadFieldU8(name='VALUE_' + str(k), description='Echoed byte'))
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='ECHO32',
                request_code=None,
                response_code=None,
                description='Echo (32 bytes)',
                request_payload=_echo32_payload,
                response_payload=_echo32_payload
            )
        )

        # ------------------------------------------------------------------------------------------------------------------
        # Append command: Utility command to check alignment
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='VALCHECK',
                request_code=None,
                response_code=None,
                description='Return a some data to check endianness',
                request_payload=[],
                response_payload=[
                    PayloadFieldU32(name='VALUE_0', description='Read as 128'),
                    PayloadFieldU32(name='VALUE_1', description='Read as 2**31 = 2147483648'),
                    PayloadFieldU16(name='VALUE_2', description='Read as 128'),
                    PayloadFieldU16(name='VALUE_3', description='Read as 2**15 = 32768'),
                    PayloadFieldU8(name='VALUE_4', description='Read as 4'),
                    PayloadFieldU8(name='VALUE_5', description='Read as 64'),
                ],
                implement=True
            )
        )

        # ------------------------------------------------------------------------------------------------------------------
        # Append command: invalid command
        self.vars.append(
            CommandDescriptor(
                firmware_config=fwcfg,
                name='INVALID',
                request_code=254,
                response_code=255,
                description='Invalid command (used in firmware, not implemented)',
                request_payload=[],
                response_payload=[],
                implement=False
            )
        )


if __name__ == '__main__':
    switch_list = SwitchList()
    analog_variable_list = AnalogVariableList()
    digital_variable_list = DigitalVariableList()
    user_pin_list = UserPinList()

    commands = CommandList(switch_list, analog_variable_list, digital_variable_list, user_pin_list)
    for command in commands.entries():
        print command.name

    print [itm.generate_enum_typename() for itm in PayloadFieldEnum.get_persistent_list().values()]

