from commands import CommandList
from entities.descriptors import CommandDescriptor, PayloadFieldEnum
from switches import SwitchList
from analogvariables import AnalogVariableList
from digitalvariables import DigitalVariableList
from userpins import UserPinList
from utils import mls
import pickle
import datetime


def generate_commandtable_header(_commands):
    assert isinstance(_commands, CommandList)  # just to help pycharm

    # read last command table version from file and update it
    command_table_version = 10
    now = datetime.datetime.now()
    generation_year = now.year
    generation_month = now.month
    generation_day = now.day
    generation_hour = now.hour
    generation_minute = now.minute
    generation_second = now.second
    # comment reading first time you execute this script...
    with open('versioning.pkl', 'r') as f:
        command_table_version = pickle.load(f)
    command_table_version
    with open('versioning.pkl', 'w') as f:
        pickle.dump(command_table_version, f)


    callback_declaration_tpl = """
// [comment]
err_t [callback](void);"""

    countdef = """
// number of defined commands
#define COMMANDS_COUNT {}
""".format(_commands.count())

    extern = """
// make command table available to anyone including this file
extern command_info_t commands_info_table[COMMANDS_COUNT];
"""

    cmdcodes_req_tpl = """#define CMDCODE_[command]_REQ [reqcode]"""
    cmdcodes_resp_tpl = """#define CMDCODE_[command]_RESP [respcode]"""

    callback_declaration_list = list()
    cmdcodes_declaration_list = list()
    for command in _commands.entries():
        assert isinstance(command, CommandDescriptor) # just to help pycharm
        callback_declaration = callback_declaration_tpl
        callback_declaration = callback_declaration.replace('[comment]', 'Callback for command {} (request code: {}), {}'.format(command.name, command.request_code, command.description))
        callback_declaration = callback_declaration.replace('[callback]', command.generate_callback_name())
        callback_declaration_list.append(callback_declaration)

        cmdcode_req_definition = cmdcodes_req_tpl
        cmdcode_req_definition = cmdcode_req_definition.replace('[command]', command.name)
        cmdcode_req_definition = cmdcode_req_definition.replace('[reqcode]', str(command.request_code))
        cmdcode_resp_definition = cmdcodes_resp_tpl
        cmdcode_resp_definition = cmdcode_resp_definition.replace('[command]', command.name)
        cmdcode_resp_definition = cmdcode_resp_definition.replace('[respcode]', str(command.response_code))
        cmdcodes_declaration_list.append(cmdcode_req_definition)
        cmdcodes_declaration_list.append(cmdcode_resp_definition)

    header_contents = ''
    header_contents += '# include "../../typedefs.h"\n'
    header_contents += countdef
    header_contents += extern
    header_contents += '\n// Command codes\n'
    header_contents += '\n'.join(cmdcodes_declaration_list)
    header_contents += '\n'
    header_contents += '\n// Callback functions\n'
    header_contents += '\n'.join(callback_declaration_list)

    # versioning information
    header_contents += '\n\n// Versioning data\n'
    header_contents += '#define COMMAND_TABLE_VERSION_NUM {}\n'.format(command_table_version)
    header_contents += '#define TEMPLATE_GENERATION_YEAR {}\n'.format(generation_year)
    header_contents += '#define TEMPLATE_GENERATION_MONTH {}\n'.format(generation_month)
    header_contents += '#define TEMPLATE_GENERATION_DAY {}\n'.format(generation_day)
    header_contents += '#define TEMPLATE_GENERATION_HOUR {}\n'.format(generation_hour)
    header_contents += '#define TEMPLATE_GENERATION_MINUTE {}\n'.format(generation_minute)
    header_contents += '#define TEMPLATE_GENERATION_SECOND {}\n'.format(generation_second)

    return header_contents


def generate_commandtable_sources(_commands):
    assert isinstance(_commands, CommandList)  # just to help pycharm

    cmdtable_tpl = """// commans info table
command_info_t commands_info_table[COMMANDS_COUNT] = {
[body]
};"""

    cmdcodes_req_tpl = """CMDCODE_[command]_REQ"""
    cmdcodes_resp_tpl = """CMDCODE_[command]_RESP"""

    cmdtable_entry_tpl = '   { .request_code = [reqcode], .response_code = [respcode], .request_payload_len = [reqlen], .callback_ptr = [callback]}'

    cmdtable_body_list = list()
    for command in _commands.entries():
        assert isinstance(command, CommandDescriptor) # just to help pycharm
        if command.implement:
            entry_src = cmdtable_entry_tpl
            entry_src = entry_src.replace('[reqcode]', mls(cmdcodes_req_tpl.replace('[command]', command.name), 43))
            entry_src = entry_src.replace('[respcode]', mls(cmdcodes_resp_tpl.replace('[command]', command.name), 44))
            entry_src = entry_src.replace('[reqlen]', mls(str(command.get_request_payload_len()), 3))
            entry_src = entry_src.replace('[callback]', mls(command.generate_callback_name(), 39))
            cmdtable_body_list.append(entry_src)

    cmdtable_src = cmdtable_tpl.replace('[body]', ',\n'.join(cmdtable_body_list))

    source_contents = ''
    source_contents += '#include "commandstable.h"\n\n'
    source_contents += cmdtable_src

    return source_contents


def generate_command_callback_template(_command):
    assert isinstance(_command, CommandDescriptor)  # just to help pycharm

    callback_declaration_tpl = """
#include "../typedefs.h"  
#include "../communication.h"  
#include "../hardware.h"

err_t [callback](void) {
[comments]
[body]
}"""

    ####################################################################################################################
    comments = ''
    comments += '    // Command: {}\n'.format(_command.name)
    comments += '    // Description: {}\n'.format(_command.description)
    comments += '    // \n'
    comments += '    // Request code: CMDCODE_{}_REQ\n'.format(_command.name)
    comments += '    // Response code: CMDCODE_{}_REQ\n'.format(_command.name)
    comments += '    // Request payload (length in bytes is {}):\n'.format(_command.get_request_payload_len())
    field_idx = 0
    for field in _command.request_payload:
        if isinstance(field, PayloadFieldEnum):
            ctype = field.generate_enum_typename()
        else:
            ctype = field.ctype
        comments += '    //     field #{} ({}) [type: {}] : {} ({})\n'.format(mls(str(field_idx), 3), str(field.length) + (' byte' if field.length == 1 else ' bytes'), mls(ctype, 10, stop_on_longer=False), mls(field.name, 10, stop_on_longer=False), field.description)
        field_idx += 1
    comments += '    // Response payload (length in bytes is {}):\n'.format(_command.get_response_payload_len())
    field_idx = 0
    for field in _command.response_payload:
        if isinstance(field, PayloadFieldEnum):
            ctype = field.generate_enum_typename()
        else:
            ctype = field.ctype
        comments += '    //     field #{} ({}) [type: {}] : {} ({})\n'.format(mls(str(field_idx), 3), str(field.length) + (' byte' if field.length == 1 else ' bytes'), mls(ctype, 10, stop_on_longer=False), mls(field.name, 10, stop_on_longer=False), field.description)
        field_idx += 1

    ####################################################################################################################
    body = ''
    body += '    // Auto-generated section: request payload fields reading\n'
    for field in _command.request_payload:
        ctype = field.generate_enum_typename() if isinstance(field, PayloadFieldEnum) else field.ctype
        body += '    {} req_{} = ({})communication_request_payload_retrieve_{}();\n'.format(ctype, field.name.lower(), ctype, field.ctype)
    for field in _command.request_payload:
        if isinstance(field, PayloadFieldEnum):
            body += '    // checking range for enum field req_{}\n'.format(field.name.lower())
            body += '    if(({})req_{} < ({}){}_min) return ERR_RANGE;\n'.format(field.ctype, field.name.lower(), field.ctype, field.generate_enum_typename())
            body += '    if(({})req_{} > ({}){}_max) return ERR_RANGE;\n'.format(field.ctype, field.name.lower(), field.ctype, field.generate_enum_typename())
    body += '    // End of auto-generated section\n\n'
    body += '    // USER CODE HERE\n'
    body += '    // TODO implement command callback function {}\n\n'.format(_command.generate_callback_name())
    for field in _command.response_payload:
        ctype = field.generate_enum_typename() if isinstance(field, PayloadFieldEnum) else field.ctype
        body += '    // TODO: assign response field {} in command {} callback\n'.format(field.name, _command.name)
        body += '    {} resp_{} = -1;\n'.format(ctype, field.name.lower())
        body += '\n'

    body += '    // Auto-generated section: response payload fields writing\n'
    body += '    communication_response_payload_appender_reset();\n'
    for field in _command.response_payload:
        ctype = field.generate_enum_typename() if isinstance(field, PayloadFieldEnum) else field.ctype
        body += '    communication_response_payload_append_{}(({})resp_{});\n'.format(field.ctype, field.ctype, field.name.lower())
    body += '    return NOERR;\n'
    body += '    // End of auto-generated section'

    callback_declaration = callback_declaration_tpl
    callback_declaration = callback_declaration.replace('[callback]', _command.generate_callback_name())
    callback_declaration = callback_declaration.replace('[comments]', comments)
    callback_declaration = callback_declaration.replace('[body]', body)

    return callback_declaration


def generate_commands_typedefs():
    enums = PayloadFieldEnum.get_persistent_list()

    enum_tpl = """
// [comment]
typedef enum {
[body]
} [typename];

#define [typename]_min [name]_[indexmin]
#define [typename]_max [name]_[indexmax]   
"""

    enum_entry_tpl = """    [prepend]_[name] = [value]"""

    contents = ''
    for enum in enums.values():

        # if enum.name == 'VARIABLE_NUMBER':
        #     print enum.name
        #     print enum.__class__.__name__
        #     print enum.index_list
        #     print enum.value_list


        enum_entries = list()
        for k in range(len(enum.index_list)):
            enum_entry_src = enum_entry_tpl
            enum_entry_src = enum_entry_src.replace('[prepend]', enum.name)
            enum_entry_src = enum_entry_src.replace('[name]', enum.value_list[k])
            enum_entry_src = enum_entry_src.replace('[value]', str(enum.index_list[k]))
            enum_entries.append(enum_entry_src)
        enum_body = ',\n'.join(enum_entries)
        enum_src = enum_tpl
        enum_src = enum_src.replace('[comment]', enum.description)
        enum_src = enum_src.replace('[typename]', enum.generate_enum_typename())
        enum_src = enum_src.replace('[name]', enum.name)
        enum_src = enum_src.replace('[indexmin]', str(enum.get_value_by_index(enum.get_index_min())))
        enum_src = enum_src.replace('[indexmax]', str(enum.get_value_by_index(enum.get_index_max())))
        enum_src = enum_src.replace('[body]', enum_body)

        contents += enum_src
    return contents


if __name__ == '__main__':
    switch_list = SwitchList()
    analog_variable_list = AnalogVariableList()
    digital_variable_list = DigitalVariableList()
    user_pin_list = UserPinList()
    commands = CommandList(switch_list, analog_variable_list, digital_variable_list, user_pin_list)
    print generate_commandtable_header(commands)
    print generate_commandtable_sources(commands)
    for command in commands.entries():
        print generate_command_callback_template(command)
    print generate_commands_typedefs()

